@@ -1,4 +1,4 @@
-from .computation_engine import apply, join, summarize, Engine, files_from, DelayedException
+from .computation_engine import apply, join, summarize, Engine, output, DelayedException
 from .fs_cache import CacheBuilder, LocalCacheBuilder
 from .std_logger import get_logger
 
