@@ -1,12 +1,9 @@
-import pdb
 import tblib.pickling_support
 # allows pickling of basic information from tracebacks:
 tblib.pickling_support.install()
 
 import abc
-import collections
 import dill
-import fnmatch
 import functools
 import itertools
 import multiprocessing
@@ -14,7 +11,6 @@ import os
 import sys
 import Queue
 import random
-import threading
 import time
 import traceback
 
@@ -144,29 +140,36 @@ class DataStream(object):
     __metaclass__ = abc.ABCMeta
 
     def __init__(self):
-        self._stream = Queue.Queue()
+        # self._stream = Queue.Queue()
+        self._streams = dict()
 
-    def empty(self):
-        return self._stream.empty()
+    def empty(self, consumer):
+        return self._streams.get(consumer).empty()
 
-    def get(self):
-        return self._stream.get()
+    def get(self, consumer):
+        return self._streams.get(consumer).get()
 
     def size(self):
         return self._size
 
     def put(self, item):
         """this is where computation results are passed to."""
-        self._stream.put(item)
+        for stream in self._streams.values():
+            stream.put(item)
 
-    def get_all_in_order(self, do_not_reraise_exceptions=False):
-        results = [self.get() for _ in range(self.size())]
-        if not do_not_reraise_exceptions:
-            for result in results:
-                if isinstance(result.value, DelayedException):
-                    result.value.re_raise()
-        results.sort(key=lambda en_item: en_item.number)
-        return [item.value for item in results]
+    def register_consumer(self, consumer):
+        assert consumer not in self._streams
+        self._streams[consumer] = Queue.Queue()
+
+    def start_computations(self):
+        """
+        starts computations at leafs first then up to the root
+        """
+        for stream in self.input_streams:
+            if True or isinstance(stream, Computation):
+                stream.start_computations()
+        self._start_computation()
+        return self
 
 
 class Source(DataStream):
@@ -174,21 +177,13 @@ class Source(DataStream):
     def __init__(self, source):
         super(Source, self).__init__()
         self._size = len(source)
-        for i, item in enumerate(source):
+        self.source = source
+        self.input_streams = []
+
+    def _start_computation(self):
+        for i, item in enumerate(self.source):
             self.put(EnumeratedItem(i, item))
 
-    def __iter__(self):
-        while not self.empty():
-            yield self.get().value
-
-
-def files_from(folder, *patterns):
-    files = []
-    for f in os.listdir(folder):
-        if not patterns or any(fnmatch.fnmatch(f, pattern) for pattern in patterns):
-            files.append(os.path.join(folder, f))
-    return Source(files)
-
 
 class Computation(DataStream):
 
@@ -207,19 +202,10 @@ class Computation(DataStream):
                     stream = Source(input_)
                 else:
                     stream = Source((input_,))
+            stream.register_consumer(self)
             streams.append(stream)
         return streams
 
-    def start_computations(self):
-        """
-        starts computations at leafs first then up to the root
-        """
-        for stream in self.input_streams:
-            if isinstance(stream, Computation):
-                stream.start_computations()
-        self._start_computation()
-        return self
-
     @abc.abstractmethod
     def compute_size(self):
         pass
@@ -229,6 +215,43 @@ class Computation(DataStream):
         pass
 
 
+class OutputNode(DataStream):
+
+    def __init__(self, input_):
+        super(OutputNode, self).__init__()
+
+        if not isinstance(input_, DataStream):
+            if isinstance(input_, (tuple, list)):
+                input_ = Source(input_)
+            else:
+                input_ = Source((input_,))
+
+        input_.register_consumer(self)
+        self.input_ = input_
+
+    def size(self):
+        return self.input_._size
+
+    def start_computations(self, do_not_reraise_exceptions=True):
+        """
+        starts computations at leafs first then up to the root
+        """
+        self.input_.start_computations()
+
+        results = [self.input_.get(self) for _ in range(self.size())]
+        if not do_not_reraise_exceptions:
+            for result in results:
+                if isinstance(result.value, DelayedException):
+                    result.value.re_raise()
+        results.sort(key=lambda en_item: en_item.number)
+        self.results = [item.value for item in results]
+
+    def get_all_in_order(self):
+        while self.results is None:
+            time.sleep(0.001)
+        return self.results
+
+
 class _UpdateManagerForJoin(object):
 
     """ manages input data from sources which might come in in arbitrary order.  handles
@@ -244,10 +267,13 @@ class _UpdateManagerForJoin(object):
         self.next_[stream_index] += 1
         # cross product over all indices of available data:
         sizes = [len(v) for v in self.data]
-        for number, perm in enumerate(itertools.product(*(range(size) for size in sizes))):
+        for perm in itertools.product(*(range(size) for size in sizes)):
             if perm not in self.seen:
                 args = tuple(self.data[j][pi] for j, pi in enumerate(perm))
                 if all(ai is not None for ai in args):
+                    number = args[0].number
+                    for ai, si in zip(args[1:], sizes[1:]):
+                        number = si * number + ai.number
                     yield number, args
                     self.seen.add(perm)
 
@@ -266,8 +292,8 @@ class JoinComputation(Computation):
         while manager.input_data_pending(self.size()):
             started_computation = False
             for i, stream in enumerate(self.input_streams):
-                if not stream.empty():
-                    item = stream.get()
+                if not stream.empty(self):
+                    item = stream.get(self)
                     if isinstance(item.value, DelayedException):
                         self.put(item)
                         return
@@ -281,7 +307,7 @@ class JoinComputation(Computation):
 class SummarizeComputation(Computation):
 
     def __init__(self, f, inputs):
-        assert len(inputs) == 1, "reduce computations only accept one input stream"
+        assert len(inputs) == 1, "summarize computations only accept one input stream"
         super(SummarizeComputation, self).__init__(f, inputs)
 
     def compute_size(self):
@@ -291,7 +317,7 @@ class SummarizeComputation(Computation):
         items = []
         stream = self.input_streams[0]
         for i in range(stream.size()):
-            item = stream.get()
+            item = stream.get(self)
             if isinstance(item.value, DelayedException):
                 self.put(item)
                 return
@@ -299,7 +325,7 @@ class SummarizeComputation(Computation):
         items.sort(key=lambda item: item.number)
         values = [v for __, v in items]
         arg = EnumeratedItem(0, values)
-        Engine.run_async(self.f, (arg,), callback=wrap_callback(0,self.put))
+        Engine.run_async(self.f, (arg,), callback=wrap_callback(0, self.put))
 
 
 class ZipComputation(Computation):
@@ -320,64 +346,75 @@ class ZipComputation(Computation):
         # sizes has size 0 means: we only had inputs of size 1
         return 1
 
-    def _collect_data_from_constant_input_streams(self):
-        constant_inputs = [None] * len(self.input_streams)
-        for i, stream in enumerate(self.input_streams):
-            if stream.size() == 1:
-                constant_inputs[i] = stream.get()
-        return constant_inputs
-
-    def _check_for_any_new_data(self):
-        for stream in self.input_streams:
-            if stream.size() > 1:
-                if not stream.empty():
-                    return True
-        return False
-
-    def _assemble_function_arguments(self, collected_inputs, seen):
-        for i, stream in enumerate(self.input_streams):
-            if not stream.empty():
-                item = stream.get()
-                collected_inputs[item.number][i] = item
-
-        for ni, arg in enumerate(collected_inputs):
-            if ni not in seen:
-                if all(c is not None for c in arg):
-                    yield ni, arg
-                    seen.add(ni)
-
     def _start_computation(self):
-        n = 0
         # wait and get the constant args which we need anyway:
         cip = self._collect_data_from_constant_input_streams()
-        for item in cip:
-            if item is not None:
-                if isinstance(item.value, DelayedException):
-                    self.put(item)
-                    return
+
+        if not self._check_and_forward_exceptions(cip):
+            return
+
         # only constant inputs ?
         if self.size() == 1:
             Engine.run_async(self.f, cip, callback=wrap_callback(0, self.put))
         else:
-            collected_inputs = [[None for __ in self.input_streams] for k in range(self.size())]
-            for i, is_ in enumerate(self.input_streams):
-                if is_.size() == 1:
-                    for ii in range(self.size()):
-                        collected_inputs[ii][i] = cip[i]
+            self._zip_all(cip)
 
-            seen = set()
-            while n < self.size():
-                avail = self._check_for_any_new_data()
-                if avail:
-                    for number, args in self._assemble_function_arguments(collected_inputs, seen):
-                        for arg in args:
-                            if isinstance(arg, DelayedException):
-                                wrap_callback(number, self.put)(arg)
-                                return
-                        Engine.run_async(self.f, args, callback=wrap_callback(number, self.put))
-                        n += 1
-                else:
-                    time.sleep(0.001)
+    def _collect_data_from_constant_input_streams(self):
+        constant_inputs = [None] * len(self.input_streams)
+        for i, stream in enumerate(self.input_streams):
+            if stream.size() == 1:
+                constant_inputs[i] = stream.get(self)
+        return constant_inputs
+
+    def _check_and_forward_exceptions(self, args):
+        for item in args:
+            if item is not None:
+                if isinstance(item.value, DelayedException):
+                    self.put(item)
+                    False
+        return True
+
+    def _zip_all(self, cip):
+        seen = set()
+        n = 0
+        input_matrix = self._setup_input_matrix(cip)
+        while n < self.size():
+            avail = self._check_for_any_new_data()
+            if avail:
+                for number, args in self._assemble_function_arguments(input_matrix, seen):
+                    if not self._check_and_forward_exceptions(args):
+                        return
+                    Engine.run_async(self.f, args, callback=wrap_callback(number, self.put))
+                    n += 1
+            else:
+                time.sleep(0.001)
+
+    def _setup_input_matrix(self, cip):
+        input_matrix = [[None for __ in self.input_streams] for k in range(self.size())]
+        for i, is_ in enumerate(self.input_streams):
+            if is_.size() == 1:
+                for ii in range(self.size()):
+                    input_matrix[ii][i] = cip[i]
+        return input_matrix
+
+    def _check_for_any_new_data(self):
+        for stream in self.input_streams:
+            if stream.size() > 1:
+                if not stream.empty(self):
+                    return True
+        return False
+
+    def _assemble_function_arguments(self, input_matrix, seen):
+        for i, stream in enumerate(self.input_streams):
+            if not stream.empty(self):
+                item = stream.get(self)
+                input_matrix[item.number][i] = item
+
+        for ni, arg in enumerate(input_matrix):
+            if ni not in seen:
+                if all(c is not None for c in arg):
+                    yield ni, arg
+                    seen.add(ni)
 
 
 def apply(inner):
@@ -404,6 +441,17 @@ def summarize(inner):
     return wrapped
 
 
+def output(inner):
+    if isinstance(inner, Computation):
+        return OutputNode(inner)
+
+    @functools.wraps(inner)
+    def wrapped(*args):
+        return OutputNode(inner(*args))
+    wrapped.inner = inner
+    return wrapped
+
+
 if __name__ == "__main__":
 
     @join
@@ -429,6 +477,7 @@ if __name__ == "__main__":
 
     r = inc(avg(inc(add(0, 1, add(inc(s1), inc(s2), inc(7))), dict(increment=2))))
 
+    r = output(r)
     r.start_computations()
 
     print(sorted(r.get_all_in_order()))
