@@ -194,26 +194,28 @@ class _CachedFunction(object):
     def clear(self):
         self._cache.clear()
 
-    def _compute_hash(self, key):
+    def _compute_hash(self, key, outer=None):
         extractor = self._lookup_hash_data_extractor_for(key)
         if extractor is not None:
             key = extractor(key)
-        if isinstance(key, basestring):
+        if isinstance(key, str):
             data = key
+        elif isinstance(key, unicode):
+            data = key.encode("utf-8")
         elif isinstance(key, (tuple, list)):
-            data = "".join(self._compute_hash(item) for item in key)
+            data = "".join(self._compute_hash(item, outer=key) for item in key)
         elif isinstance(key, set):
-            data = "".join(self._compute_hash(item) for item in sorted(key))
+            data = "".join(self._compute_hash(item, outer=key) for item in sorted(key))
         elif isinstance(key, dict):
-            data = "".join(self._compute_hash(item) for item in key.items())
+            data = "".join(self._compute_hash(item, outer=key) for item in key.items())
         elif hasattr(key, "__dict__"):
-            data = self._compute_hash(key.__dict__)
+            data = self._compute_hash(key.__dict__, outer=key)
         elif isinstance(key, (bool, int, long, float,)):
             data = str(self._pickler(key))
         elif key is None:
             data = "__None__"
         else:
-            raise Exception("can not compute hash for %r" % (key,))
+            raise Exception("can not compute hash for %r contained in %r" % (key, outer))
         if not isinstance(data, basestring):
             raise RuntimeError("implementation error: data should be str, but is %s" % type(data))
         muncher = hashlib.sha1()
